<?php

namespace Drupal\pagedesigner;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityFieldManager;
use Drupal\Core\Language\LanguageManager;
use Drupal\Core\Routing\CurrentRouteMatch;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\pagedesigner\Entity\Element;
use Drupal\pagedesigner\Service\ElementHandler;

/**
 * Base class for handling operations on elements and content entities.
 *
 * @todo Refactor class to split responsibilities and reduce dependencies.
 */
class PagedesignerService implements PagedesignerServiceInterface {
  use StringTranslationTrait;

  /**
   * The output of the last operation.
   *
   * @var mixed
   */
  protected $output = NULL;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManager
   */
  protected $languageManager = NULL;

  /**
   * The element handler.
   *
   * @var \Drupal\pagedesigner\Service\ElementHandler
   */
  protected $elementHandler = NULL;

  /**
   * The current route.
   *
   * @var \Drupal\Core\Routing\CurrentRouteMatch
   */
  protected $currentRoute = NULL;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser = NULL;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManager
   */
  protected $entityFieldManager = NULL;

  /**
   * Cache the edit access.
   *
   * @var bool
   */
  protected $editAccess = NULL;

  /**
   * Create a new instance.
   *
   *   The element handler.
   *   The language manager.
   *   The current route.
   *   The current user.
   *   The entity field manager.
   */
  public function __construct(ElementHandler $element_handler, LanguageManager $language_manager, CurrentRouteMatch $current_route, AccountProxyInterface $current_user, EntityFieldManager $entity_field_manager) {
    $this->elementHandler = $element_handler;
    $this->languageManager = $language_manager;
    $this->currentRoute = $current_route;
    $this->currentUser = $current_user;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * Returns the output of the last operation.
   *
   * @return mixed
   *   The output of the last operation.
   */
  public function getOutput() {
    return $this->output;
  }

  /**
   * Get all pagedesigner fields of the entity.
   *
   *   The entity to get the fields from.
   *
   * @return array
   *   The list of fields, keyed by field name.
   */
  public function getPagedesignerFields(ContentEntityInterface $entity) {
    $pagedesignerFields = [];
    $fields = $entity->getFields(TRUE);
    foreach ($fields as $field) {
      if ($field->getFieldDefinition()->getType() == 'pagedesigner_item') {
        $pagedesignerFields[$field->getName()] = $field;
      }
    }
    return $pagedesignerFields;
  }

  /**
   * Adds the container value to the given entity.
   *
   * If no container exists, a container will be created.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to add the containers to.
   *
   * @return bool
   *   Whether the entity has been changed.
   */
  public function addContainer(ContentEntityInterface &$entity) {
    $changed = FALSE;
    $fields = $this->getPagedesignerFields($entity);
    foreach ($fields as $field) {
      if ($this->addContainerForField($entity, $field->getName())) {
        $changed = TRUE;
      }
    }
    return $changed;
  }

  /**
   * {@inheritdoc}
   */
  public function addContainerForField(ContentEntityInterface &$entity, string $field_name) {
    $changed = FALSE;
    $language = $entity->language()->getId();
    if ($entity->get($field_name)->isEmpty()) {
      if (!$entity->isDefaultTranslation()) {
        $container = $this->getContainer($entity->getUntranslated(), $field_name);
        if (!$container->hasTranslation($language)) {
          $container->addTranslation($language);
          $container->save();
        }
      }
      else {
        $container = $this->getContainer($entity, $field_name);
      }
      $entity->get($field_name)->entity = $container;
      $changed = TRUE;
    }
    elseif (!$entity->isDefaultTranslation()) {
      $container = $entity->get($field_name)->entity;
      if (!$container->hasTranslation($language)) {
        $container->addTranslation($language);
        $container->save();
      }
    }
    return $changed;
  }

  /**
   * Replaces the container linked in the entity.
   *
   *   The entity to replace the container in.
   */
  public function overrideContainers(ContentEntityInterface &$entity) {
    $fields = $this->getPagedesignerFields($entity);
    foreach ($fields as $field) {
      $container = $this->createContainer($entity->language()->getId(), $entity);
      $entity->getUntranslated()->get($field->getName())->entity = $container;
    }
    foreach ($entity->getTranslationLanguages(FALSE) as $language) {
      foreach ($fields as $field) {
        $entity->getTranslation($language->getId())->get($field->getName())->entity = $entity->getUntranslated()->get($field->getName())->entity;
      }
    }
  }

  /**
   * Saves the containers in all fields.
   *
   *   The entity to save the container on.
   */
  public function saveContainer(ContentEntityInterface &$entity) {
    $fields = $this->getPagedesignerFields($entity);
    foreach ($fields as $field) {
      if (!$entity->get($field->getName())->isEmpty()) {
        $container = $entity->get($field->getName())->entity;
        $container->entity = $entity;
        $container->save();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function enableTranslation(ContentEntityInterface &$entity, string $field_name = NULL) {
    foreach ($entity->getTranslationLanguages(FALSE) as $language) {
      if ($entity->hasTranslation($language->getId())) {
        $entity = $entity->getTranslation($language->getId());
        if (!empty($field_name)) {
          if ($this->addContainerForField($entity, $field_name)) {
            $entity->save();
          }
        }
        else {
          if ($this->addContainer($entity)) {
            $entity->save();
          }
        }
      }
    }
  }

  /**
   * Loads/creates the container for the given entity in the default language.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being rendered.
   * @param string $field_name
   *   The pagedesigner field name.
   *
   * @return \Drupal\pagedesigner\Entity\Element
   *   The container.
   */
  public function getContainer(ContentEntityInterface $entity, $field_name = 'field_pagedesigner_content') {
    $container = $entity->get($field_name)->entity;
    $language = $entity->language()->getId();
    if ($container == NULL) {
      $container = $this->createContainer($language, $entity);
    }
    else {
      $container = $this->getLanguage($container, $language);
    }
    return $container;
  }

  /**
   * Create a new container.
   *
   * @param string $language
   *   The langcode for the container.
   * @param Drupal\pagedesigner\ContentEntityInterface $entity
   *   The entity for which to create the container (optional).
   *
   * @return \Drupal\pagedesigner\Entity\Element
   *   The container.
   */
  public function createContainer(string $language, ContentEntityInterface $entity = NULL) {
    $container = Element::create(['type' => 'container', 'name' => 'container']);
    $container->langcode->value = $language;
    if (!empty($entity) && !$entity->isNew()) {
      $container->entity->entity = $entity;
    }
    $container->save();
    return $container;
  }

  /**
   * {@inheritdoc}
   */
  public function clearPagedesigner(ContentEntityInterface &$entity, string $field_name = NULL) {
    if (empty($field)) {
      $fields = $this->getPagedesignerFields($entity);
      foreach ($fields as $field_name => $field) {
        $this->doClearPagedesigner($entity, $field_name);
      }
    }
    else {
      $this->doClearPagedesigner($entity, $field_name);
    }
  }

  /**
   * Mark pagedesigner elements as deleted on a field in an entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to clear on.
   * @param string $field_name
   *   The field to clear.
   */
  protected function doClearPagedesigner(ContentEntityInterface &$entity, string $field_name = NULL) {
    if (empty($field_name)) {
      return;
    }
    foreach ($entity->getTranslationLanguages(TRUE) as $language) {
      if ($entity->hasTranslation($language->getId())) {
        $entity = $entity->getTranslation($language->getId());
        $container = $this->getContainer($entity, $field_name);
        if (!empty($container)) {
          $this->elementHandler->delete($container);
        }
      }
    }
  }

  /**
   * Gets or creates the translation of the given container.
   *
   * @param \Drupal\pagedesigner\Entity\Element $container
   *   The container.
   * @param string $language
   *   The language.
   *
   * @return \Drupal\pagedesigner\Entity\Element
   *   The container in the given language.
   */
  public function getLanguage(Element $container, $language) {
    if (!$container->hasTranslation($language)) {
      $container->addTranslation($language);
      $container->save();
    }
    return $container->getTranslation($language);
  }

  /**
   * Whether assets should be loaded for the pagedesigner.
   *
   * Checks, if the pagedesigner route is active
   * and the entity (current or given) is editable
   * by the user (current or given).
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity being loaded.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   The user loading the entity.
   *
   * @return bool
   *   True if assets should be loaded, false otherwise.
   */
  public function isEditMode(ContentEntityInterface $entity = NULL, AccountInterface $user = NULL) {
    return $this->isPagedesignerRoute() && $this->hasEditAccess($entity, $user);
  }

  /**
   * Whether the current route is the pagedesigner edit route.
   *
   * @return bool
   *   True if the pagedesigner edit route is active, false otherwise.
   */
  public function isPagedesignerRoute() {
    return $this->currentRoute->getRouteName() == 'pagedesigner.node.edit_mode';
  }

  /**
   * Returns true if the given entity may be edited by the user.
   *
   *   The entity to check.
   *   The user to check access for.
   *
   * @return bool
   *   True if the user has access to the entity, false otherwise.
   */
  public function hasEditAccess(ContentEntityInterface $entity = NULL, AccountInterface $user = NULL) {
    if ($this->editAccess !== NULL) {
      return $this->editAccess;
    }
    if ($entity == NULL) {
      if ($this->isPagedesignerRoute()) {
        $entity = $this->currentRoute->getParameter('node');
      }
      else {
        return FALSE;
      }
    }
    if ($user == NULL) {
      $user = $this->currentUser;
    }

    $hook_result = TRUE;
    $access_results = \Drupal::moduleHandler()->invokeAll('pagedesigner_access', [$entity, $user]);
    /** @var \Drupal\Core\Access\AccessResult $access_result */
    foreach ($access_results as $access_result) {
      if (!$access_result->isAllowed()) {
        $hook_result = FALSE;
      }
    }

    $this->editAccess = $user->hasPermission('edit pagedesigner element entities')
    && $entity->access('update', $user)
    && $hook_result;
    return $this->editAccess;

  }

  /**
   * Get the element handler.
   *
   * @return Drupal\pagedesigner\Service\ElementHandler
   *   Return the pagedesigner element handler.
   */
  protected function getHandler() {
    return $this->elementHandler;
  }

}
