<?php

namespace Drupal\Tests\pagedesigner\Kernel\HandlerTests;

use Drupal\pagedesigner\Entity\Element;
use Drupal\Tests\pagedesigner\Kernel\AbstractPagedesignerTest;
use Drupal\ui_patterns\Definition\PatternDefinition;

/**
 * Base class to test a field handler with a predefined entity defintion.
 *
 * Tests serializing, getting and rendering of the entity.
 *
 * @group pagedesigner
 */
abstract class HandlerTestBase extends AbstractPagedesignerTest {
  /**
   * The element handler.
   *
   * @var \Drupal\pagedesigner\Plugin\pagedesigner\HandlerPluginInterface
   */
  protected $handlerId = '';

  /**
   * The element handler.
   *
   * @var \Drupal\pagedesigner\Plugin\pagedesigner\HandlerPluginInterface
   */
  protected $handler = NULL;

  /**
   * The handler manager.
   *
   * @var \Drupal\pagedesigner\Service\HandlerPluginManager
   */
  protected $handlerManager = NULL;

  /**
   * The entity definition.
   *
   * @var array
   */
  protected $entityDefinition = [];

  /**
   * The field definitions.
   *
   * @var array
   */
  protected $fieldDefinition = [];

  /**
   * The field data.
   *
   * @var array
   */
  protected $sampleDate = [];

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();
    $this->handlerManager = \Drupal::service('plugin.manager.pagedesigner_handler');
  }

  /**
   * Test the handler with the entity.
   */
  public function testHandler() {
    $this->entity = Element::create($this->entityDefinition);
    $this->entity->setHandlerType($this->handlerId);
    $this->handler = $this->handlerManager->createInstance($this->handlerId);
    $this->assertTrue($this->entity != NULL);
    $this->assertTrue($this->handler != NULL);
    $this->collectPatternsTest();
    $this->collectAttachmentsTest();
    $this->adaptPatternsTest();
    $this->serializeTest();
    $this->getTest();
    $this->renderTest();
    $this->renderForEditTest();
  }

  /**
   * Test the collect patterns function of the handler.
   *
   * @return array
   *   The collected patterns.
   */
  protected function collectPatternsTest() {
    $patterns = [];
    $this->handler->collectPatterns($patterns);
    $this->assertTrue(\is_array($patterns));
    foreach ($patterns as $pattern) {
      $this->assertTrue($pattern instanceof PatternDefinition);
    }
    return $patterns;
  }

  /**
   * Test the collect attachments function of the handler.
   *
   * @return array
   *   The collected attachments.
   */
  protected function collectAttachmentsTest() {
    $attachments = ['library' => []];
    $this->handler->collectAttachments($attachments);
    $this->assertTrue(\is_array($attachments));
    $this->assertTrue(\is_array($attachments['library']));
    return $attachments;
  }

  /**
   * Test the adapt patterns function of the handler.
   *
   * @return array
   *   The adapted patterns.
   */
  protected function adaptPatternsTest() {
    $patterns = [];
    $this->handler->adaptPatterns($patterns);
    $this->assertTrue(\is_array($patterns));
    return $patterns;
  }

  /**
   * Test the serialize function of the handler.
   *
   * @return array
   *   The produced serialized representation of the entity.
   */
  protected function serializeTest() {
    $result = [];
    $this->handler->serialize($this->entity, $result);
    $this->assertTrue(\is_array($result));
    return $result;
  }

  /**
   * Test the get function of the handler.
   *
   * @return string
   *   The produced text representation of the entity.
   */
  protected function getTest() {
    $value = '';
    $this->handler->get($this->entity, $value);
    $this->assertTrue(\is_scalar($value));
    return $value;
  }

  /**
   * Test the render function of the handler.
   *
   * @return array
   *   The produced render representation of the entity.
   */
  protected function renderTest() {
    $build = [];
    $this->handler->render($this->entity, $build);
    $this->assertTrue(\is_array($build));
    return $build;
  }

  /**
   * Test the render function of the handler.
   *
   * @return array
   *   The produced render representation of the entity.
   */
  protected function renderForEditTest() {
    $build = [];
    $this->handler->renderForEdit($this->entity, $build);
    $this->assertTrue(\is_array($build));
    return $build;
  }

}
